# SPDX-License-Identifier: GPL-3.0-or-later
import hashlib
import logging

log = logging.getLogger(__name__)
import io
import os.path

_cache = {}


def snapshot(path):
    """
        A snapshot of the contents of an external file.

        The special value NO_SUCH_FILE is returned when path does not
        refer to an existing external file. However, an exception is
        raised if an existing file vanishes between two calls.

        The implementation trusts the operating system about modification
        times, and assumes that an unchanged time stamp implies unchanged
        contents. Malicious or unadvised users may change timestamps.

        Moreover, an overwrite will not be detected if it is more recent
        than the smallest interval representable by operating timestamps.

        The implementation relies on the MD5 hash to detect modified
        contents. For such a non-cryptographic use,  the probability of
        collision (2^-64) can be neglected for all practical needs.
    """
    # We expect some files to be sources in many contexts, like the
    # main .tex document. In order to spare some checksum
    # computations, we cache the result.

    # Distinct paths refering to the same external file should be
    # rare, so we do not attempt to detect them.
    try:
        c, t = _cache[path]
    except KeyError:
        c, t = None, None

    if os.path.exists(path):
        mtime = os.path.getmtime(path)
        if c is None:
            log.debug('%s contents are now watched', path)
            c = _checksum_algorithm(path)
            t = mtime
        elif c == NO_SUCH_FILE:
            log.debug('%s has been created', path)
            c = _checksum_algorithm(path)
            t = mtime
        elif t == mtime:
            log.debug('%s has the same mtime', path)
        else:
            assert t < mtime, 'mtime decreased: ' + path
            t = mtime
            checksum = _checksum_algorithm(path)
            if checksum == c:
                log.debug('%s rewritten with same checksum', path)
            else:
                log.debug('%s rewritten with new contents.', path)
                c = checksum
    elif c is None:
        log.debug('%s will be watched once created', path)
        c = NO_SUCH_FILE
    else:
        assert c == NO_SUCH_FILE, path + ' vanished'
        log.debug('%s does not exist yet', path)

    _cache[path] = (c, t)
    return c


# Md5 values are represented as bytes. None is used above.
NO_SUCH_FILE = bytes()


def _checksum_algorithm(path):
    with open(path, 'br') as stream:
        result = hashlib.md5()
        while True:
            data = stream.read(io.DEFAULT_BUFFER_SIZE)
            if not data:
                return result.digest()
            result.update(data)


# These two functions encapsulate the hexadecimal representation of
# checksums other than NO_SUCH_FILE.  In order to ease formatting, all
# results are guaranteed to have a common length.
cs_str_len = 32


def cs2str(checksum):
    if checksum == NO_SUCH_FILE:
        result = 'No such file                    '
    else:
        result = ''.join('{:02X}'.format(byte) for byte in checksum)
    assert len(result) == cs_str_len
    return result


def str2cs(string):
    assert len(string) == cs_str_len
    if string == 'No such file                    ':
        return NO_SUCH_FILE
    else:
        return bytes(int(string[i:i + 2], base=16) for i in range(0, len(string), 2))


# Manual tests

# import time
# logging.basicConfig (level = logging.DEBUG)
# t = 'tmpfile'
# if os.path.exists (t):
#     os.remove (t)

# time.sleep (0.1)
# s0 = snapshot (t)
# assert snapshot (t) == s0
# assert s0 == NO_SUCH_FILE

# print ('writing foo')
# with open (t, 'w') as f:
#     f.write ('foo')

# time.sleep (0.1)
# s1 = snapshot (t)
# assert s1 != NO_SUCH_FILE

# time.sleep (0.1)
# assert snapshot (t) == s1

# print ('writing bar')
# with open (t, 'w') as f:
#     f.write ('bar')

# time.sleep (0.1)
# s2 = snapshot (t)
# assert s2 != NO_SUCH_FILE
# assert s2 != s1

# time.sleep (0.1)
# assert snapshot (t) == s2

# print ('writing bar')
# with open (t, 'w') as f:
#     f.write ('bar')

# time.sleep (0.1)
# assert snapshot (t) == s2

# os.remove (t)
# print ('OK')



# SPDX-License-Identifier: GPL-3.0-or-later
# (c) Emmanuel Beffara, 2008
"""
Support for nomenclatures with package 'nomencl'.

This module simply wraps the functionality of the 'index' module with default
values for the 'nomencl' package.
"""

import rubber.index
import rubber.module_interface


class Module(rubber.module_interface.Module):

    def __init__(self, document, opt):
        self.dep = rubber.index.Index(document, 'nlo', 'nls', 'ilg')
        self.dep.style = 'nomencl.ist'



# SPDX-License-Identifier: GPL-3.0-or-later
import rubber.depend
import rubber.module_interface


class Module(rubber.module_interface.Module):

    def __init__(self, document, opt):
        job = document.basename()

        document.add_product(job + '.ist')

        dep = rubber.depend.Shell(('makeglossaries', job))
        dep.add_source(job + '.aux')

        if opt is None or 'nomain' not in opt:
            glo = job + '.glo'
            # FIXME: does probably fail with --inplace and friends.
            document.add_source(glo)

            dep.add_product(glo)
            dep.add_product(job + '.gls')
            dep.add_product(job + '.glg')

        if opt is not None and 'acronym' in opt:
            acn = job + '.acn'
            document.add_source(acn)

            dep.add_product(acn)
            dep.add_product(job + '.acr')
            dep.add_product(job + '.alg')



# SPDX-License-Identifier: GPL-3.0-or-later
# (c) Emmanuel Beffara, 2002--2006
# (c) Sebastian Kapfer, 2015
# vim: noet:ts=4
"""
This is the command line interface for Rubber.
"""

import argparse
import os.path
import sys
import shutil
import tempfile
# bzip2 and/or gzip may be imported depending on command line options.
import rubber.converters.compressor
import rubber.converters.latex
import rubber.converters.literate
import rubber.depend
import rubber.environment
from rubber.util import _
import logging

msg = logging.getLogger(__name__)
import rubber.util

try:
    import rubber.version
    # pylint: disable=E1101
    RUBBER_VERSION = rubber.version.version
except ImportError:
    RUBBER_VERSION = "unknown"

# The expected entry point is the main procedure, with one of these
# three values to track the command name (which may differ from
# sys.argv [0] for symbolic links).
RUBBER_PLAIN = 0
RUBBER_PIPE = 1
RUBBER_INFO = 2


def parse_opts(command_name):

    class DeprecatedAction(argparse.Action):

        def __call__(self, parser, namespace, values, option_string=None):
            raise rubber.SyntaxError('obsolete option: ' + option_string)

    if command_name == RUBBER_PLAIN:
        parser = argparse.ArgumentParser(description='Run TeX until a document is built.')
        parser.add_argument('source', nargs='+')
        mode = parser.add_mutually_exclusive_group()
        mode.add_argument('--clean',
                          action='store_true',
                          help='remove produced files instead of compiling')
    elif command_name == RUBBER_PIPE:
        parser = argparse.ArgumentParser(
            description='Build a TeX document received on standard input.')
    else:  # command_name == RUBBER_INFO
        parser = argparse.ArgumentParser(description="""
            Filter messages from the log created by a previous run of rubber.
            One of the following options must be selected:
            boxes, check (the default), deps, errors,refs, rules or warning.
        """)
        parser.add_argument('source', nargs='+')
        parser.set_defaults(info_action='check')
        info_action = parser.add_mutually_exclusive_group()
        info_action.add_argument('--boxes',
                                 action='store_const',
                                 dest='info_action',
                                 const='boxes',
                                 help='report overfull and underfull boxes')
        info_action.add_argument('--check',
                                 action='store_const',
                                 dest='info_action',
                                 const='check',
                                 help='report errors or warnings (default action)')
        info_action.add_argument('--deps',
                                 action='store_const',
                                 dest='info_action',
                                 const='deps',
                                 help="show the target file's dependencies")
        info_action.add_argument('--errors',
                                 action='store_const',
                                 dest='info_action',
                                 const='errors',
                                 help='show all errors that occured during compilation')
        info_action.add_argument('--refs',
                                 action='store_const',
                                 dest='info_action',
                                 const='refs',
                                 help='show the list of undefined references')
        info_action.add_argument('--rules',
                                 action='store_const',
                                 dest='info_action',
                                 const='rules',
                                 help='print the dependency rules including intermediate results')
        info_action.add_argument('--warnings',
                                 action='store_const',
                                 dest='info_action',
                                 const='warnings',
                                 help='show all LaTeX warnings')

    parser.set_defaults(
        epilogue=[],
        place='.',
        prologue=[],
    )
    compress = parser.add_mutually_exclusive_group()
    place = parser.add_mutually_exclusive_group()

    # Non-mode options, sorted by short name, else by long name.

    compress.add_argument('-b',
                          '--bzip2',
                          action='store_const',
                          const='bzip2',
                          dest='compress',
                          help='compress the final document with bzip2')

    parser.add_argument('-c',
                        '--command',
                        action='append',
                        dest='prologue',
                        metavar='CMD',
                        help='run the directive CMD before parsing')

    class PDFAction(argparse.Action):

        def __call__(self, parser, namespace, values, option_string=None):
            if 'module dvips' in namespace.epilogue:
                namespace.epilogue.append('module ps2pdf')
            else:
                namespace.prologue.append('module pdftex')

    parser.add_argument('-d',
                        '--pdf',
                        action=PDFAction,
                        nargs=0,
                        help="shortcut for -c 'module pdftex' or -e 'module ps2pdf'")

    parser.add_argument('-e',
                        '--epilogue',
                        action='append',
                        metavar='CMD',
                        help='run the directive CMD after parsing')

    if command_name == RUBBER_PLAIN:
        mode.add_argument('-f',
                          '--force',
                          action='store_true',
                          help='force at least one compilation')

    if command_name != RUBBER_PIPE:
        place.add_argument('--inplace',
                           action='store_const',
                           dest='place',
                           const=None,
                           help='compile the documents from their source directory')

    place.add_argument('--into',
                       dest='place',
                       metavar='DIR',
                       help='go to directory DIR before compiling')

    parser.set_defaults(texpath=[])
    parser.add_argument('-I',
                        '--texpath',
                        action='append',
                        metavar='DIR',
                        help='add DIR to the search path for LaTeX')

    parser.add_argument('--jobname', help='set the job name for the first target')

    if command_name == RUBBER_PIPE:
        parser.add_argument('-k',
                            '--keep',
                            action='store_true',
                            help='keep the temporary files after compiling')

    parser.add_argument('-l',
                        '--landscape',
                        nargs=0,
                        action=DeprecatedAction,
                        help='obsolete option, must not be used')

    class ModuleAction(argparse.Action):

        def __call__(self, parser, namespace, values, option_string=None):
            namespace.prologue.append('module ' + values.replace(':', ' ', 1))

    parser.add_argument('-m',
                        '--module',
                        action=ModuleAction,
                        metavar='MOD[:OPTS]',
                        help="shortcut for -c 'module MOD OPTS'")

    parser.add_argument('-n',
                        '--maxerr',
                        type=int,
                        default=10,
                        metavar='NUM',
                        help='display at most NUM errors (default %(default)i)')

    class PostAction(argparse.Action):

        def __call__(self, parser, namespace, values, option_string=None):
            namespace.epilogue.append('module ' + values.replace(':', ' ', 1))

    if command_name != RUBBER_INFO:
        parser.add_argument('-o',
                            '--post',
                            action=PostAction,
                            metavar='MOD[:OPTS]',
                            help="shortcut for -e 'module MOD OPTS'")

    parser.add_argument('--only',
                        metavar='SOURCE[,SOURCE,...]',
                        help='only include the specified SOURCES')

    parser.add_argument('-p',
                        '--ps',
                        action='append_const',
                        dest='epilogue',
                        const='module dvips',
                        help="shortcut for -e 'module dvips'")

    parser.add_argument('-q',
                        '--quiet',
                        action='count',
                        help='decrease verbosity (may be repeated)')

    class ReadAction(argparse.Action):

        def __call__(self, parser, namespace, values, option_string=None):
            namespace.prologue.append('read ' + values)

    parser.add_argument('-r',
                        '--read',
                        action=ReadAction,
                        metavar='FILE',
                        help="shortcut for -c 'read FILE'")

    parser.add_argument('--readopts',
                        action=DeprecatedAction,
                        help='obsolete option, must not be used')

    parser.add_argument('-s',
                        '--short',
                        action='store_true',
                        help='display errors in a compact form')

    parser.add_argument('-S',
                        '--src-specials',
                        action='append_const',
                        dest='prologue',
                        const='set src-specials yes',
                        help="shortcut for -c 'set src-specials yes'")

    parser.add_argument('--synctex',
                        action='append_const',
                        dest='prologue',
                        const='synctex',
                        help='shortcut for -c synctex')

    parser.add_argument('--unsafe',
                        '--shell-escape',
                        action='store_true',
                        help='permits the document to run external commands')

    parser.add_argument('-v',
                        '--verbose',
                        action='count',
                        help='increase verbosity (may be repeated)')

    parser.add_argument('--version', action='version', version='%(prog)s ' + RUBBER_VERSION)

    warn_values = ('all', 'boxes', 'misc', 'refs')

    class WarnAction(argparse.Action):

        def __call__(self, parser, namespace, values, option_string=None):
            if values == 'all':
                namespace.warn_boxes = True
                namespace.warn_misc = True
                namespace.warn_refs = True
            else:
                setattr(namespace, 'warn_' + values, True)

    if command_name != RUBBER_INFO:
        parser.set_defaults(warn_boxes=False, warn_misc=False, warn_refs=False)
        parser.add_argument('-W',
                            '--warn',
                            action=WarnAction,
                            metavar='TYPE',
                            choices=warn_values,
                            help='report warnings matching TYPE: ' + ','.join(warn_values))

    compress.add_argument('-z',
                          '--gzip',
                          action='store_const',
                          const='gzip',
                          dest='compress',
                          help='compress the final document with gzip')

    args = parser.parse_args()

    if command_name == RUBBER_PLAIN and args.clean \
       and (args.warn_boxes or args.warn_refs or args.warn_misc):
        raise rubber.SyntaxError('incompatible options: --clean and --warn')

    loggingFormat = "%(message)s"
    logLevel = logging.INFO
    if args.verbose:
        logLevel -= 10 * args.verbose
        if args.verbose > 1:
            loggingFormat = "%(module)s:%(funcName)s:%(message)s"
    if args.quiet:
        logLevel += 10 * args.quiet
    if logging.ERROR < logLevel:
        logLevel = logging.ERROR
    if logLevel < logging.DEBUG:
        logLevel = logging.DEBUG
    logging.basicConfig(level=logLevel, format=loggingFormat)

    return args


def prepare_source(filename, command_name, env, options):
    """
    Prepare the dependency node for the main LaTeX run.
    Returns the filename of the main LaTeX source file, which might
    change for various reasons (adding a .tex suffix; preprocessors;
    pipe dumping).
    When this is done, the file must exist on disk, otherwise this
    function must exit(1) or exit(2).
    """
    path = rubber.util.find_resource(filename, suffix=".tex")

    if path is None:
        raise rubber.GenericError(_("Main document not found: '%s'") % filename)

    base, ext = os.path.splitext(path)

    if ext in rubber.converters.literate.literate_preprocessors.keys():
        src = base + ".tex"
        # FIXME kill src_node
        src_node = rubber.converters.literate.literate_preprocessors[ext](src, path)
        if command_name == RUBBER_PLAIN and not options.clean:
            if not options.unsafe:
                raise rubber.SyntaxError(_("Running external commands requires --unsafe."))
            # Produce the source from its dependency rules, if needed.
            try:
                src_node.make()
            except rubber.depend.MakeError as e:
                raise rubber.GenericError(
                    _("Producing the main LaTeX file {} failed: {}")
                    .format(src, e.msg))
    else:
        src = path

    env.final = env.main = rubber.converters.latex.LaTeXDep(env, src, options.jobname)

    return src


def main(command_name):
    assert command_name in (RUBBER_PLAIN, RUBBER_PIPE, RUBBER_INFO)

    try:
        options = parse_opts(command_name)

        msg.debug(_("This is Rubber version %s.") % RUBBER_VERSION)

        if command_name == RUBBER_PIPE:
            # Generate a temporary source file, and pretend it has
            # been given on the command line.
            args = (prepare_source_pipe(),)
        else:
            args = options.source

        if options.place is None:  # --inplace
            # Compute all absolute paths before the first chdir.
            args = map(os.path.abspath, args)
        elif options.place != '.':  # non default --into
            msg.info(_("changing to %s from --into option") % options.place)
            # Make arguments relative to the new directory,
            # go there then proceed normally.
            args = map(lambda p: os.path.relpath(p, options.place), args)
            try:
                os.chdir(options.place)
            except OSError as e:
                raise rubber.GenericError \
                    (_("Error changing to %s from --into option: %s") \
                     % (options.place, e.strerror))

        for src in args:

            msg.debug(_("about to process file '%s'") % src)

            if options.place is None:  # --inplace
                # Chdir to the absolute path, then keep the base name.
                src_dirname, src = os.path.split(src)
                try:
                    os.chdir(src_dirname)
                except OSError as e:
                    raise rubber.GenericError \
                        (_("Error changing to directory %s for %s: %s")\
                         % (src_dirname, src, e.strerror))

            # prepare the source file.  this may require a pre-processing
            # step, or dumping stdin.  thus, the input filename may change.
            # in case of build mode, preprocessors will be run as part of
            # prepare_source.
            env = rubber.environment.Environment()
            src = prepare_source(src, command_name, env, options)

            # safe mode is off during the prologue
            env.is_in_unsafe_mode_ = True

            if options.only is not None:
                env.main.includeonly(options.only)

            # at this point, the LaTeX source file must exist; if it is
            # the result of pre-processing, this has happened already.
            # the main LaTeX file is not found via find_file (unlike
            # most other resources) by design:  paths etc may be set up
            # from within via rubber directives, so that wouldn't make a
            # whole lot of sense.
            if not os.path.exists(src):
                raise rubber.GenericError(_("LaTeX source file not found: '%s'") % src)

            env.path.extend(options.texpath)

            saved_vars = env.main.vars.copy()
            for cmd in options.prologue:
                cmd = rubber.util.parse_line(cmd, env.main.vars)
                env.main.command(cmd[0], cmd[1:], {'file': 'command line'})
            env.main.vars = saved_vars

            # safe mode is enforced for anything that comes from the .tex file
            env.is_in_unsafe_mode_ = options.unsafe

            env.main.parse()

            saved_vars = env.main.vars.copy()
            for cmd in options.epilogue:
                cmd = rubber.util.parse_line(cmd, env.main.vars)
                env.main.command(cmd[0], cmd[1:], {'file': 'command line'})
            env.main.vars = saved_vars

            if options.compress is not None:
                last_node = env.final
                filename = last_node.primary_product()
                if options.compress == 'gzip':
                    import gzip
                    env.final = rubber.converters.compressor.Node(gzip.GzipFile, '.gz', filename)
                else:
                    assert options.compress == 'bzip2'
                    import bz2
                    env.final = rubber.converters.compressor.Node(bz2.BZ2File, '.bz2', filename)

            if command_name == RUBBER_PIPE:
                process_source_pipe(env, src, options)
            elif command_name == RUBBER_INFO:
                process_source_info(env, options.info_action, options.short)
            elif options.clean:
                for node in env.final.all_producers():
                    node.clean()
                cache_path = env.main.basename('.rubbercache')
                if os.path.exists(cache_path):
                    msg.debug(_("removing %s"), cache_path)
                    os.remove(cache_path)
            else:
                build(options, RUBBER_PLAIN, env)

        if (command_name == RUBBER_PLAIN and options.clean) \
           or (command_name == RUBBER_PIPE and not options.keep):
            rubber.depend.clean_all_products()

    except KeyboardInterrupt:
        msg.warning(_("*** interrupted"))
        sys.exit(1)
    except rubber.SyntaxError as e:
        print('error: ' + str(e), file=sys.stderr)
        sys.exit(1)
    except rubber.GenericError as e:
        print('error: ' + str(e), file=sys.stderr)
        sys.exit(2)


def build(options, command_name, env):
    """
    Build the final product.
    """
    assert command_name == RUBBER_PIPE \
            or (command_name == RUBBER_PLAIN and not options.clean)

    cache_path = env.main.basename('.rubbercache')
    if os.path.exists(cache_path):
        if command_name == RUBBER_PLAIN and options.force:
            msg.debug(_('Ignoring cache file if any because of --force.'))
        else:
            rubber.depend.load_cache(cache_path)

    try:
        if command_name == RUBBER_PLAIN and options.force:
            ret = env.main.make()
            if env.final is not env.main:
                ret = env.final.make() or ret
        else:
            ret = env.final.make()
    except rubber.depend.MakeError as e:
        msg.info(_("There were errors compiling %s: %s."),
                 env.main.source(), e.msg)
        number = options.maxerr
        for err in e.errors:
            if number == 0:
                msg.info(_("More errors."))
                break
            display(options.short, **err)
            number -= 1
        # Ensure a message even with -q.
        raise rubber.GenericError(_("Stopping because of compilation errors."))

    if ret:
        rubber.depend.save_cache(cache_path, env.final)
    else:
        msg.info(_("nothing to be done for %s"), env.main.source())

    if options.warn_boxes or options.warn_misc or options.warn_refs:
        # FIXME
        log = env.main.log
        if not env.main.parse_log():
            msg.error(_("cannot read the log file"))
            return 1
        for err in log.parse(boxes=options.warn_boxes,
                             refs=options.warn_refs,
                             warnings=options.warn_misc):
            display(options.short, **err)


def display(short, kind, text, **info):
    """
    Print an error or warning message. The argument 'kind' indicates the
    kind of message, among "error", "warning", "abort", the argument
    'text' is the main text of the message, the other arguments provide
    additional information, including the location of the error.
    """
    if kind == "error":
        if text[0:13] == "LaTeX Error: ":
            text = text[13:]
        msg.warning(rubber.util._format(info, text))
        if "code" in info and info["code"] and not short:
            if "macro" in info:
                del info["macro"]
            msg.warning(rubber.util._format(info, _("leading text: ") + info["code"]))
    elif kind == "abort":
        if short:
            m = _("compilation aborted ") + info["why"]
        else:
            m = _("compilation aborted: %s %s") % (text, info["why"])
        msg.warning(rubber.util._format(info, m))
    else:
        assert kind == "warning"
        msg.warning(rubber.util._format(info, text))


def prepare_source_pipe():
    """
    Dump the standard input in a file, and set up that file
    the same way we would normally process LaTeX sources.
    """

    # FIXME: with a better program structure, the context manager
    # should remove the input file.

    try:
        # Make a temporary on-disk copy of the standard input,
        # in the current working directory.
        # The name will have the form "rubtmpXXX.tex.
        with tempfile.NamedTemporaryFile(suffix='.tex', prefix='rubtmp', dir='.',
                                         delete=False) as srcfile:
            # note the tempfile name so we can remove it later
            pipe_tempfile = srcfile.name
            # copy stdin into the tempfile
            msg.info(_("saving the input in %s") % pipe_tempfile)
            shutil.copyfileobj(sys.stdin.buffer, srcfile)
    except IOError:
        raise rubber.GenericError(_("cannot create temporary file for the main LaTeX source"))

    return pipe_tempfile


def process_source_pipe(env, pipe_tempfile, options):
    """
    Build the document, and dump the result on stdout.
    """
    try:
        build(options, RUBBER_PIPE, env)
        filename = env.final.primary_product()
        try:
            # dump the results on standard output
            with open(filename, "rb") as output:
                shutil.copyfileobj(output, sys.stdout.buffer)
        except IOError:
            raise rubber.GenericError(_("error copying the product '%s' to stdout") % filename)
    finally:
        # clean the intermediate files
        if not options.keep:
            for node in env.final.all_producers():
                node.clean()
            cache_path = env.main.basename('.rubbercache')
            if os.path.exists(cache_path):
                msg.debug(_("removing %s"), cache_path)
                os.remove(cache_path)
            if os.path.exists(pipe_tempfile):
                msg.info(_("removing %s"), pipe_tempfile)
                os.remove(pipe_tempfile)


def process_source_info(env, act, short):
    if act == "deps":
        print(" ".join(env.final.all_leaves()))

    elif act == "rules":
        for node in env.final.all_producers():
            print("\n%s:" % " ".join(node.products()))
            print(" ".join(node.sources))

    else:
        # Check for a log file and extract information from it if it exists,
        # accroding to the argument's value.
        log = env.main.log
        if not env.main.parse_log():
            raise rubber.GenericError(_("Parsing the log file failed"))

        if act == "boxes":
            for err in log.get_boxes():
                display(short, **err)
            else:
                msg.info(_("There is no bad box."))
        elif act == "check":
            finished = False
            for err in log.get_errors():
                display(short, **err)
                finished = True
            if finished:
                return 0
            msg.info(_("There was no error."))
            for err in log.get_references():
                display(short, **err)
                finished = True
            if finished:
                return 0
            msg.info(_("There is no undefined reference."))
            for err in log.get_warnings():
                display(short, **err)
            else:
                msg.info(_("There is no warning."))
            for err in log.get_boxes():
                display(short, **err)
            else:
                msg.info(_("There is no bad box."))
        elif act == "errors":
            for err in log.get_errors():
                display(short, **err)
            else:
                msg.info(_("There was no error."))
        elif act == "refs":
            for err in log.get_references():
                display(short, **err)
            else:
                msg.info(_("There is no undefined reference."))
        else:
            assert act == "warnings"
            for err in log.get_warnings():
                display(short, **err)
            else:
                msg.info(_("There is no warning."))


def main_rubber_plain():
    main(RUBBER_PLAIN)


def main_rubber_info():
    main(RUBBER_INFO)


def main_rubber_pipe():
    main(RUBBER_PIPE)


def main_rubber_lsmod():
    from rubber.util import iter_rubber_modules
    for module in iter_rubber_modules():
        print(module)



# SPDX-License-Identifier: GPL-3.0-or-later
# (c) Emmanuel Beffara, 2004--2006
# vim: noet:ts=4
"""
Common support for makeindex and xindy external tools.
"""

import os
import logging

msg = logging.getLogger(__name__)
import rubber.depend
from rubber.util import _

# FIXME: this class may probably be simplified a lot if inheriting
# from rubber.depend.Shell instead of rubber.depend.Node.


class Index(rubber.depend.Node):
    """
    This class represents a single index.
    """

    def __init__(self, doc, source, target, transcript):
        """
        Initialize the index, by specifying the source file (generated by
        LaTeX), the target file (the output of makeindex) and the transcript
        (e.g. .ilg) file.  Transcript is used by glosstex.py.
        """
        super().__init__()
        src = doc.basename(with_suffix="." + source)
        tgt = doc.basename(with_suffix="." + target)
        log = doc.basename(with_suffix="." + transcript)
        doc.add_product(src)
        self.add_product(tgt)
        self.add_product(log)
        self.add_source(src)
        doc.add_source(tgt)
        self.doc = doc
        self.cmd = ["makeindex", src, "-q", "-o", tgt, "-t", log]
        self.lang = None  # only for xindy
        self.modules = []  # only for xindy
        self.opts = []
        self.path = []
        self.style = None  # only for makeindex
        self.command_env = None

    def do_language(self, args):
        if len(args) != 1:
            raise rubber.SyntaxError(_("invalid syntax for directive '%s'") % args)
        lang = args[0]
        self.lang = lang

    def do_modules(self, args):
        self.modules.extend(args)

    def do_order(self, args):
        for opt in args:
            if opt == "standard":
                self.opts = []
            elif opt == "german":
                self.opts.append("-g")
            elif opt == "letter":
                self.opts.append("-l")
            else:                msg.warning(
                    _("unknown option '%s' for 'makeidx.order'") % opt)

    def do_path(self, args):
        if len(args) != 1:
            raise rubber.SyntaxError(_("invalid syntax for directive '%s'") % "path")
        path = args[0]
        self.path.append(path)

    def do_style(self, args):
        if len(args) != 1:
            raise rubber.SyntaxError(_("invalid syntax for directive '%s'") % "style")
        style = args[0]
        self.style = style

    def do_tool(self, args):
        if len(args) != 1:
            raise rubber.SyntaxError(_("invalid syntax for directive '%s'") % "tool")
        tool = args[0]
        if tool not in ("makeindex", "xindy"):
            msg.error(_("unknown indexing tool '%s'") % tool)
        self.cmd[0] = tool

    def run(self):
        # No more settings are expected, we compute the
        # command once and for all.
        if self.command_env == None:
            if self.cmd[0] == "makeindex":
                self.cmd.extend(self.opts)
                if self.style:
                    self.cmd.append('-s')
                    self.cmd.append(self.style)
                path_var = "INDEXSTYLE"
            else:  # self.cmd [0] == "texindy"
                for opt in self.opts:
                    if opt == "-g":
                        if self.lang != "":
                            msg.warning(_("'language' overrides 'order german'"))
                        else:
                            self.lang = "german-din"
                    else:  # opt == "-l"
                        self.modules.append("letter-ordering")
                        msg.warning(_("use 'module letter-ordering' instead of 'order letter'"))
                for mod in self.modules:
                    self.cmd.append('--module')
                    self.cmd.append(mod)
                if self.lang:
                    self.cmd.append('--language')
                    self.cmd.append(self.lang)
                path_var = "XINDY_SEARCHPATH"

            if self.path != []:
                self.command_env = {path_var: ':'.join(self.path + [os.getenv(path_var, '')])}
            else:
                self.command_env = {}

        # The actual run.
        return rubber.util.execute(self.cmd, env=self.command_env) == 0



# SPDX-License-Identifier: GPL-3.0-or-later
# (c) Emmanuel Beffara, 2003--2006
"""
Epsfig support for Rubber.

The package 'epsfig' is a somewhat deprecated interface to the graphics module
that simply provides a macro \\psfig with a keyval list argument to include
an EPS figure file.
"""

from rubber.util import parse_keyval
import rubber.module_interface


class Module(rubber.module_interface.Module):

    def __init__(self, document, opt):
        document.modules.register('graphics')
        _, self.hook_includegraphics = document.hooks['includegraphics']
        # We proceed as if \epsfbox and \includegraphics were equivalent.
        document.hook_macro('epsfbox', 'oa', self.hook_epsfbox)
        document.hook_macro('epsffile', 'oa', self.hook_epsfbox)
        document.hook_macro('epsfig', 'a', self.hook_epsfig)
        document.hook_macro('psfig', 'a', self.hook_epsfig)

    def hook_epsfbox(self, loc, optional, argument):
        self.hook_includegraphics(loc, False, optional, argument)

    def hook_epsfig(self, loc, argument):
        # We just translate this into an equivalent call to \includegraphics.
        options = parse_keyval(argument)
        if 'file' not in options:
            return
        self.hook_includegraphics(loc, False, argument, options['file'])



# SPDX-License-Identifier: GPL-3.0-or-later
import rubber.module_interface


class Module(rubber.module_interface.Module):

    def __init__(self, document, opt):

        nav = document.basename(with_suffix='.nav')
        document.add_product(nav)
        document.watch_file(nav)

        snm = document.basename(with_suffix='.snm')
        document.add_product(snm)
        document.watch_file(snm)

        toc = document.basename(with_suffix='.toc')
        document.add_product(toc)
        document.watch_file(toc)

        document.modules.register('hyperref')



# SPDX-License-Identifier: GPL-3.0-or-later
# (c) Emmanuel Beffara, 2002--2006
# vim:noet:ts=4
"""
This module contains utility functions and classes used by the main system and
by the modules for various tasks.
"""

import os.path, stat
import logging

msg = logging.getLogger(__name__)
import re
from string import whitespace
import pkgutil
import subprocess

import rubber.latex_modules

#-- Message writers --{{{1


# The function `_' is defined here to prepare for internationalization.
def _(txt):
    return txt


def _format(where, text):
    """
    Format the given text into a proper error message, with file and line
    information in the standard format. Position information is taken from
    the dictionary given as first argument.
    """
    if where is None or where == {}:
        return text

    if "file" in where and where["file"] is not None:
        pos = where["file"]
        if "line" in where and where["line"]:
            pos = "%s:%d" % (pos, int(where["line"]))
            if "last" in where:
                if where["last"] != where["line"]:
                    pos = "%s-%d" % (pos, int(where["last"]))
        pos = pos + ": "
    else:
        pos = ""
    if "macro" in where:
        text = "%s (in macro %s)" % (text, where["macro"])
    if "page" in where:
        text = "%s (page %d)" % (text, int(where["page"]))
    if "pkg" in where:
        text = "[%s] %s" % (where["pkg"], text)
    return pos + text


#-- Keyval parsing --{{{1

re_keyval = re.compile("\
[ ,]*\
(?P<key>[^ \t\n{}=,]+)\
([ \n\t]*=[ \n\t]*\
(?P<val>({|[^{},]*)))?")


def parse_keyval(str):
    """
    Parse a list of 'key=value' pairs, with the syntax used in LaTeX's
    standard 'keyval' package. The value returned is simply a dictionary that
    contains all definitions found in the string. For keys without a value,
    the dictionary associates the value None.
    If str is None, consider it as empty.
    """
    dict = {}
    while str:
        m = re_keyval.match(str)
        if not m:
            break
        d = m.groupdict()
        str = str[m.end():]
        if not d["val"]:
            dict[d["key"]] = None
        elif d["val"] == '{':
            val, str = match_brace(str)
            dict[d["key"]] = val
        else:
            dict[d["key"]] = d["val"].strip()
    return dict


def match_brace(str):
    """
    Split the string at the first closing brace such that the extracted prefix
    is balanced with respect to braces. The return value is a pair. If the
    adequate closing brace is found, the pair contains the prefix before the
    brace and the suffix after the brace (not containing the brace). If no
    adequate brace is found, return the whole string as prefix and an empty
    string as suffix.
    """
    level = 0
    for pos in range(0, len(str)):
        if str[pos] == '{':
            level = level + 1
        elif str[pos] == '}':
            level = level - 1
            if level == -1:
                return (str[:pos], str[pos + 1:])
    return (str, "")


#-- Checking for program availability --{{{1

checked_progs = {}


def prog_available(prog):
    """
    Test whether the specified program is available in the current path, and
    return its actual path if it is found, or None.
    """
    pathsep = ";" if os.name == "nt" else ":"
    fileext = ".exe" if os.name == "nt" else ""
    if prog in checked_progs:
        return checked_progs[prog]
    for path in os.getenv("PATH").split(pathsep):
        file = os.path.join(path, prog) + fileext
        if os.path.exists(file):
            st = os.stat(file)
            if stat.S_ISREG(st.st_mode) and (st.st_mode & 0o111):
                checked_progs[prog] = file
                return file
    checked_progs[prog] = None
    return None


#-- Parsing commands --{{{1

re_variable = re.compile("(?P<name>[a-zA-Z]+)")


def parse_line(line, dict):
    """
    Decompose a string into a list of elements. The elements are separated by
    spaces, single and double quotes allow escaping of spaces (and quotes).
    Elements can contain variable references with the syntax '$VAR' (with only
    letters in the name) or '${VAR}'.

    If the argument 'dict' is defined, it is considered as a hash containing
    the values of the variables. If it is None, elements with variables are
    replaced by sequences of litteral strings or names, as follows:
        parse_line(" foo  bar${xy}quux toto  ")
            --> ["foo", ["'bar", "$xy", "'quux"], "toto"]
    """
    elems = []
    i = 0
    size = len(line)
    while i < size:
        while i < size and line[i] in whitespace:
            i = i + 1
        if i == size:
            break

        open = 0  # which quote is open
        arg = ""  # the current argument, so far
        if not dict:
            composed = None  # the current composed argument

        while i < size:
            c = line[i]

            # Open or close quotes.

            if c in '\'\"':
                if open == c:
                    open = 0
                elif open:
                    arg = arg + c
                else:
                    open = c

            # '$' introduces a variable name, except within single quotes.

            elif c == '$' and open != "'":

                # Make the argument composed, if relevant.

                if not dict:
                    if not composed:
                        composed = []
                    if arg != "":
                        composed.append("'" + arg)
                    arg = ""

                # Parse the variable name.

                if i + 1 < size and line[i + 1] == '{':
                    end = line.find('}', i + 2)
                    if end < 0:
                        name = line[i + 2:]
                        i = size
                    else:
                        name = line[i + 2:end]
                        i = end + 1
                else:
                    m = re_variable.match(line, i + 1)
                    if m:
                        name = m.group("name")
                        i = m.end()
                    else:
                        name = ""
                        i = i + 1

                # Append the variable or its name.

                if dict:
                    if name in dict:
                        arg = arg + str(dict[name])
                    elif name in (
                            'cwd',
                            'base',
                            'ext',
                            'path',
                            'latex',
                            'program',
                            'engine',
                            'file',
                            'line',
                    ):
                        msg.error(_('Obsolete variable: ' + name))
                    elif name in (
                            'graphics_suffixes',
                            'src-specials',
                            'logfile_limit',
                    ):
                        msg.error(_('Write-only variable: ' + name))
                    else:
                        msg.error(_('Unknown variable: ' + name))

                else:
                    composed.append("$" + name)
                continue

            # Handle spaces.

            elif c in whitespace:
                if open:
                    arg = arg + c
                else:
                    break
            else:
                arg = arg + c
            i = i + 1

        # Append the new argument.

        if dict or not composed:
            elems.append(arg)
        else:
            if arg != "":
                composed.append("'" + arg)
            elems.append(composed)

    return elems


def explode_path(name="PATH"):
    """
    Parse an environment variable into a list of paths, and return it as an array.
    """
    path = os.getenv(name)
    if path is not None:
        return path.split(":")
    else:
        return []


def find_resource(name, suffix="", paths=[]):
    """
    find the indicated file, mimicking what latex would do:
    tries adding a suffix such as ".bib", or looking in the specified paths.
    if unsuccessful, returns None.
    """
    name = name.strip()

    if os.path.exists(name):
        return name
    elif suffix != "" and os.path.exists(name + suffix):
        return name + suffix

    for path in paths:
        fullname = os.path.join(path, name)
        if os.path.exists(fullname):
            return fullname
        elif suffix != "" and os.path.exists(fullname + suffix):
            return fullname + suffix

    return None


def execute(prog, env={}, pwd=None, out=None):
    """
    Silently execute an external program. The `prog' argument is the list
    of arguments for the program, `prog[0]' is the program name. The `env'
    argument is a dictionary with definitions that should be added to the
    environment when running the program. The standard output is passed
    line by line to the `out' function (or discarded by default).
    """
    msg.info(_("executing: %s") % " ".join(prog))
    if pwd:
        msg.debug(_("  in directory %s") % pwd)
    if env != {}:
        msg.debug(_("  with environment: %r") % env)

    progname = prog_available(prog[0])
    if not progname:
        msg.error(_("%s not found") % prog[0])
        return 1

    penv = os.environ.copy()
    for (key, val) in env.items():
        penv[key] = val

    with subprocess.Popen(prog,
                          executable=progname,
                          env=penv,
                          cwd=pwd,
                          stdin=subprocess.DEVNULL,
                          stdout=subprocess.PIPE,
                          stderr=None) as process:

        if out is not None:
            for line in process.stdout:
                out(line)
        else:
            process.stdout.readlines()

        ret = process.wait()
        msg.debug(_("process %d (%s) returned %d") % (process.pid, prog[0], ret))
        return ret


def iter_rubber_modules():
    package = rubber.latex_modules
    prefix = package.__name__ + "."

    for _, modname, _ in pkgutil.iter_modules(package.__path__, prefix):
        modname_components = modname.split(".")
        last_component = modname_components[-1]
        yield last_component



# SPDX-License-Identifier: GPL-3.0-or-later
import rubber.module_interface


class Module(rubber.module_interface.Module):

    def __init__(self, document, opt):
        document.program = 'xelatex'
        document.engine = 'XeLaTeX'
        document.register_post_processor(old_suffix='.pdf', new_suffix='.pdf')
